package com.jiuqi.sql.impl;

import java.io.OutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Calendar;
import java.util.Date;
import com.jiuqi.util.JqLib;

public class RW
{
    private static final int dayOfYear = 365;
    private static final int dayOf4Year = 1461;
    private static final int dayOfCentury = 36524;
    private static final int dayOf4Century = 146097;
    private static final int[][] monthDays;
    private static final int hoursPerDay = 24;
    private static final int minsPerHour = 60;
    private static final int secsPerMin = 60;
    private static final int mSecsPerSec = 1000;
    private static final int minsPerDay = 1440;
    private static final int secsPerDay = 86400;
    private static final int mSecsPerDay = 86400000;
    
    static {
        monthDays = new int[][] { { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }, { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 } };
    }
    
    private static int[] decodeDelphiDateFully(final double dateTime) {
        final int[] result = new int[5];
        int t = (int)dateTime + 693594;
        if (t > 0) {
            result[3] = t % 7 + 1;
            --t;
            int y;
            for (y = 1; t >= 146097; t -= 146097, y += 400) {}
            int i = t / 36524;
            int d = t % 36524;
            if (i == 4) {
                --i;
                d += 36524;
            }
            y += i * 100;
            i = d / 1461;
            d %= 1461;
            y += i * 4;
            i = d / 365;
            d %= 365;
            if (i == 4) {
                --i;
                d += 365;
            }
            y += i;
            result[4] = ((y % 4 == 0 && (y % 100 != 0 || y % 400 == 0)) ? 1 : 0);
            final int[] dayTable = RW.monthDays[result[4]];
            int m = 1;
            while (true) {
                i = dayTable[m - 1];
                if (d < i) {
                    break;
                }
                d -= i;
                ++m;
            }
            result[0] = y;
            result[1] = m - 1;
            result[2] = d + 1;
        }
        return result;
    }
    
    private static double encodeDelphiDateFully(final int year, final int dayOfYear) {
        int dayCount = (year - 1 - 1899) * 365;
        int leapYearCount = 0;
        for (int i = 1900; i <= year - 1; ++i) {
            if (JqLib.isLeapYear(i)) {
                ++leapYearCount;
            }
        }
        dayCount += leapYearCount;
        return ++dayCount + dayOfYear;
    }
    
    private static int[] decodeDelphiTime(final double dateTime) {
        final int[] result = new int[4];
        final int time = (int)(8.64E7 * (dateTime - (int)dateTime));
        final int minCount = time / 60000;
        final int mSecCount = time % 60000;
        result[0] = minCount / 60;
        result[1] = minCount % 60;
        result[2] = mSecCount / 1000;
        result[3] = mSecCount % 1000;
        return result;
    }
    
    private static double encodeDelphiTime(final int hour, final int minute, final int second, final int msec) {
        final int minCount = 60 * hour + minute;
        final int mSecCount = 1000 * second + msec;
        final int time = 60000 * minCount + mSecCount;
        return time / 8.64E7;
    }
    
    private static Date decodeDelphiDateTime(final double dateTime) {
        final int[] date = decodeDelphiDateFully(dateTime);
        final int[] time = decodeDelphiTime(dateTime);
        final Calendar calendar = Calendar.getInstance();
        calendar.setTimeInMillis(0L);
        calendar.set(date[0], date[1], date[2], time[0], time[1], time[2]);
        return calendar.getTime();
    }
    
    private static double encodeDelphiDateTime(final Date date) {
        final Calendar calendar = Calendar.getInstance();
        calendar.setTimeInMillis(0L);
        calendar.setTime(date);
        final int year = calendar.get(1);
        final int dayOfYear = calendar.get(6);
        final int hour = calendar.get(10);
        final int minute = calendar.get(12);
        final int sec = calendar.get(13);
        final int mSec = calendar.get(14);
        return encodeDelphiDateFully(year, dayOfYear) + encodeDelphiTime(hour, minute, sec, mSec);
    }
    
    public static byte read(final InputStream stream) throws IOException {
        return (byte)stream.read();
    }
    
    public static void write(final OutputStream stream, final byte value) throws IOException {
        stream.write(value);
    }
    
    public static boolean readBoolean(final InputStream stream) throws IOException {
        return stream.read() > 0;
    }
    
    public static void writeBoolean(final OutputStream stream, final boolean value) throws IOException {
        stream.write(value ? 1 : 0);
    }
    
    public static boolean readBoolean(final byte[] data, final int index) {
        return data[index] == 84 || data[index] == 116 || data[index] == 49;
    }
    
    public static void writeBoolean(final byte[] data, final int index, final boolean value) {
        if (value) {
            data[index] = 84;
        }
        else {
            data[index] = 70;
        }
    }
    
    public static String readString(final InputStream stream, final int len) throws IOException {
        final byte[] str = new byte[len];
        stream.read(str, 0, len);
        return bytesToString(str);
    }
    
    public static void writeString(final OutputStream stream, final String value, final int len) throws IOException {
        int l = 0;
        if (value != null) {
            final byte[] str = value.getBytes();
            stream.write(str, 0, (str.length > len) ? len : str.length);
            if (str.length > len) {
                return;
            }
            l = str.length;
        }
        for (int i = 1; i <= len - l; ++i) {
            stream.write(0);
        }
    }
    
    private static String bytesToString(final byte[] data) {
        if (data == null || data.length == 0) {
            return null;
        }
        for (int i = data.length - 1; i >= 0; --i) {
            if (data[i] != 0 && data[i] != 32) {
                return new String(data, 0, i + 1);
            }
        }
        return "";
    }
    
    public static String readString(final byte[] data, final int index, final int len) {
        final byte[] str = new byte[len];
        System.arraycopy(data, index, str, 0, len);
        return bytesToString(str);
    }
    
    public static void writeString(final byte[] data, final int index, final String value, final int len) {
        final byte[] dataToWrite = value.getBytes();
        System.arraycopy(dataToWrite, 0, data, index, (dataToWrite.length > len) ? len : dataToWrite.length);
        if (len > dataToWrite.length) {
            final byte[] restData = new byte[len - dataToWrite.length];
            final int ind = dataToWrite.length + index;
            System.arraycopy(restData, 0, data, ind, restData.length);
        }
    }
    
    public static Date readDateTime(final byte[] data, final int index) {
        final double datetime = readDouble(data, index);
        return decodeDelphiDateTime(datetime);
    }
    
    public static Date readDateTime(final InputStream stream) throws IOException {
        final double datetime = readDouble(stream);
        return decodeDelphiDateTime(datetime);
    }
    
    public static void writeDateTime(final byte[] data, final int index, final Date dateValue) {
        final double datetime = encodeDelphiDateTime(dateValue);
        writeDouble(data, index, datetime);
    }
    
    public static void writeDateTime(final OutputStream stream, final Date dateValue) throws IOException {
        final double datetime = encodeDelphiDateTime(dateValue);
        writeDouble(stream, datetime);
    }
    
    public static void writeDateTime(final OutputStream stream, final long dateValue) throws IOException {
        writeLong(stream, dateValue);
    }
    
    public static void writeDateTime(final byte[] data, final int index, final long dateValue) {
        writeLong(data, index, dateValue);
    }
    
    public static double readDouble(final InputStream stream) throws IOException {
        return Double.longBitsToDouble(readLong(stream));
    }
    
    public static void writeDouble(final OutputStream stream, final double value) throws IOException {
        writeLong(stream, Double.doubleToLongBits(value));
    }
    
    public static double readDouble(final byte[] data, final int index) {
        return Double.longBitsToDouble(readLong(data, index));
    }
    
    public static void writeDouble(final byte[] data, final int index, final double value) {
        writeLong(data, index, Double.doubleToLongBits(value));
    }
    
    public static float readFloat(final InputStream stream) throws IOException {
        return Float.intBitsToFloat(readInt(stream));
    }
    
    public static void writeFloat(final OutputStream stream, final float value) throws IOException {
        writeInt(stream, Float.floatToIntBits(value));
    }
    
    public static float readFloat(final byte[] data, final int index) {
        return Float.intBitsToFloat(readInt(data, index));
    }
    
    public static void writeFloat(final byte[] data, final int index, final float value) {
        writeInt(data, index, Float.floatToIntBits(value));
    }
    
    public static int readInt(final InputStream stream) throws IOException {
        final int b1 = stream.read() & 0xFF;
        final int b2 = stream.read() & 0xFF;
        final int b3 = stream.read() & 0xFF;
        final int b4 = stream.read() & 0xFF;
        final int result = (b4 << 0) + (b3 << 8) + (b2 << 16) + (b1 << 24);
        return result ^ Integer.MIN_VALUE;
    }
    
    public static void writeInt(final OutputStream stream, int value) throws IOException {
        value ^= Integer.MIN_VALUE;
        final byte[] b = { (byte)(value >> 24 & 0xFF), (byte)(value >> 16 & 0xFF), (byte)(value >> 8 & 0xFF), (byte)(value >> 0) };
        stream.write(b);
    }
    
    public static int readInt(final byte[] data, final int index) {
        final int b1 = data[index] & 0xFF;
        final int b2 = data[index + 1] & 0xFF;
        final int b3 = data[index + 2] & 0xFF;
        final int b4 = data[index + 3] & 0xFF;
        final int result = (b4 << 0) + (b3 << 8) + (b2 << 16) + (b1 << 24);
        return result ^ Integer.MIN_VALUE;
    }
    
    public static void writeInt(final byte[] data, final int index, int value) {
        value ^= Integer.MIN_VALUE;
        data[index + 3] = (byte)(value >> 0);
        data[index + 2] = (byte)(value >> 8 & 0xFF);
        data[index + 1] = (byte)(value >> 16 & 0xFF);
        data[index + 0] = (byte)(value >> 24 & 0xFF);
    }
    
    public static long readLong(final InputStream stream) throws IOException {
        final long b1 = stream.read() & 0xFF;
        final long b2 = stream.read() & 0xFF;
        final long b3 = stream.read() & 0xFF;
        final long b4 = stream.read() & 0xFF;
        final long b5 = stream.read() & 0xFF;
        final long b6 = stream.read() & 0xFF;
        final long b7 = stream.read() & 0xFF;
        final long b8 = stream.read();
        return (b8 << 56) + (b7 << 48) + (b6 << 40) + (b5 << 32) + (b4 << 24) + (b3 << 16) + (b2 << 8) + (b1 << 0);
    }
    
    public static void writeLong(final OutputStream stream, final long value) throws IOException {
        final byte[] b = { (byte)(value >> 0), (byte)(value >> 8 & 0xFFL), (byte)(value >> 16 & 0xFFL), (byte)(value >> 24 & 0xFFL), (byte)(value >> 32 & 0xFFL), (byte)(value >> 40 & 0xFFL), (byte)(value >> 48 & 0xFFL), (byte)(value >> 56 & 0xFFL) };
        stream.write(b);
    }
    
    public static long readLong(final byte[] data, final int index) {
        final long b1 = data[index] & 0xFF;
        final long b2 = data[index + 1] & 0xFF;
        final long b3 = data[index + 2] & 0xFF;
        final long b4 = data[index + 3] & 0xFF;
        final long b5 = data[index + 4] & 0xFF;
        final long b6 = data[index + 5] & 0xFF;
        final long b7 = data[index + 6] & 0xFF;
        final long b8 = data[index + 7];
        return (b8 << 56) + (b7 << 48) + (b6 << 40) + (b5 << 32) + (b4 << 24) + (b3 << 16) + (b2 << 8) + (b1 << 0);
    }
    
    public static void writeLong(final byte[] data, final int index, final long value) {
        data[index] = (byte)(value >> 0);
        data[index + 1] = (byte)(value >> 8 & 0xFFL);
        data[index + 2] = (byte)(value >> 16 & 0xFFL);
        data[index + 3] = (byte)(value >> 24 & 0xFFL);
        data[index + 4] = (byte)(value >> 32 & 0xFFL);
        data[index + 5] = (byte)(value >> 40 & 0xFFL);
        data[index + 6] = (byte)(value >> 48 & 0xFFL);
        data[index + 7] = (byte)(value >> 56 & 0xFFL);
    }
    
    public static int readWord(final InputStream stream) throws IOException {
        final int b1 = stream.read() & 0xFF;
        final int b2 = stream.read() & 0xFF;
        return (b2 << 8) + (b1 << 0);
    }
    
    public static void writeWord(final OutputStream stream, final int value) throws IOException {
        final byte[] b = { (byte)value, (byte)(value >> 8 & 0xFF) };
        stream.write(b);
    }
    
    public static int readWord(final byte[] data, final int index) {
        final int b1 = data[index] & 0xFF;
        final int b2 = data[index + 1] & 0xFF;
        return (b2 << 8) + (b1 << 0);
    }
    
    public static void writeWord(final byte[] data, final int index, final int value) {
        data[index] = (byte)(value >> 0);
        data[index + 1] = (byte)(value >> 8 & 0xFF);
    }
    
    public static long readDWord(final InputStream stream) throws IOException {
        final int b1 = stream.read() & 0xFF;
        final int b2 = stream.read() & 0xFF;
        final int b3 = stream.read() & 0xFF;
        final int b4 = stream.read() & 0xFF;
        return (b4 << 24) + (b3 << 16) + (b2 << 8) + (b1 << 0);
    }
    
    public static void writeDWord(final OutputStream stream, final long value) throws IOException {
        final byte[] b = { (byte)(value >> 0), (byte)(value >> 8 & 0xFFL), (byte)(value >> 16 & 0xFFL), (byte)(value >> 24 & 0xFFL) };
        stream.write(b);
    }
    
    public static long readDWord(final byte[] data, final int index) {
        final int b1 = data[index] & 0xFF;
        final int b2 = data[index + 1] & 0xFF;
        final int b3 = data[index + 2] & 0xFF;
        final int b4 = data[index + 3] & 0xFF;
        return (b4 << 24) + (b3 << 16) + (b2 << 8) + (b1 << 0);
    }
    
    public static void writeDWord(final byte[] data, final int index, final long value) {
        data[index] = (byte)(value >> 0);
        data[index + 1] = (byte)(value >> 8 & 0xFFL);
        data[index + 2] = (byte)(value >> 16 & 0xFFL);
        data[index + 3] = (byte)(value >> 24 & 0xFFL);
    }
}
